from itertools import islice
from typing import List

from rich.text import Text

from .renderables.logrecord import LevelName
from .renderables.logrecord import LogRecord
from .widgets.levels import Levels

DEFAULT_NUM_LINES = 100
MAX_NUM_LINES = 100_000


# KeyValueLoader is a class that loads log files that have a key-value format.
# The following key-value pairs are expected in this order:
#
# level=<LEVEL>
# ts=<DATE TIME> in the format %Y-%m-%dT%H:%M:%S,%f e.g. 2022-04-08T10:52:20,371211
# process=<process name>
# process_id=<process ID>
# caller=<module>:<lineno>
# msg="<message>"

class KeyValueLoader:
    def __init__(self, filename: str):
        self.filename = filename
        self._lines: List[str] = ...
        """The original lines read from the log file."""
        self._size = 0
        """The number of lines in the log file."""
        self._records = []
        """Processed lines"""
        self._offset = 0
        """The line number of the first record, i.e. which line in the log file."""

    def load(self):
        """Loads the complete log file in a list of strings."""
        with open(self.filename, 'r') as fd:
            self._lines = fd.read().split('\n')

        self._size = len(self._lines)

    def size(self) -> int:
        """Returns the total number of lines in the log file."""
        return self._size

    def process(self, start: int = 0, num_lines: int = DEFAULT_NUM_LINES, levels: Levels = None):
        """Process a number of lines and creates a list of Records for those lines."""

        # * could keep track of those line that have been processed -> no need to process again
        # * sub_messages are e.g. Traceback or multiline messages

        self._offset = start

        sub_message: List[str] = []
        in_sub_message = False
        records = []
        match_count = 0
        for count, line in enumerate(islice(self._lines, start, None)):
            if count > MAX_NUM_LINES:
                break
            if not line.startswith("level="):
                in_sub_message = True
                sub_message.append(line)
                continue
            if in_sub_message:
                # print('\n'.join(sub_message))
                sub_message = []
                in_sub_message = False

            level, ts, process, process_id, caller, msg = line.split(maxsplit=5)
            level = LevelName[level[6:]].value

            if levels is None or levels.is_on(level):

                record = LogRecord(
                    level=level,
                    ts=ts[3:],
                    process=process[8:],
                    process_id=process_id[11:],
                    caller=caller[7:],
                    msg=msg[4:].strip('"')  # also removes the double quotes around the message
                )

                records.append(record)
                match_count += 1

            if match_count >= num_lines:
                break

        # if sub_message:
        #     print('\n'.join(sub_message))

        self._records = records

    def __str__(self):
        return "\n".join(self._records)

    def reprocess(self, start: int, num_lines: int, levels: Levels):
        # start and stop are line numbers of the log file

        # item is the correct mapping of the requested lines in the self._records list.
        #
        # if not 0 < start - offset < len(self._records) -> reprocess the lines
        # if not 0 < start + num_lines - offset < len(self._records) -> reprocess the lines

        nr_records = len(self._records)

        if not 0 <= start - self._offset < nr_records or \
                not 0 < start + num_lines - self._offset <= nr_records:
            self.process(start, num_lines, levels)

    def get_records(self,
                    start: int = 0,
                    num_lines: int = DEFAULT_NUM_LINES,
                    levels: Levels = None) -> List[LogRecord]:

        self.process(start, num_lines, levels)

        return self._records

    def get_text(self, start: int = 0, num_lines: int = DEFAULT_NUM_LINES, levels: Levels = None) -> Text:

        self.process(start, num_lines, levels)

        record: LogRecord
        text = Text()
        [
            text.append(record.__rich__()).append('\n')
            for record in self._records
        ]
        return text


if __name__ == "__main__":

    fn = '/Users/rik/Desktop/general.log'
    fn = '/Users/rik/data/CSL/log/general.log.2022-04-08'
    fn = '/Users/rik/data/CSL/log/general.log'

    loader = KeyValueLoader(fn)
    loader.load()
    loader.process(0, 20, None)

    import rich
    rich.print(loader.get_text(0, 20))
    rich.print("---")
    rich.print(loader.get_text(2, 4))
    rich.print("---")
    rich.print(loader.get_text(18, 2))
    rich.print("---")
    rich.print(loader.get_text(20, 2))



from __future__ import annotations

from typing import List
from typing import Optional

from rich.console import ConsoleRenderable
from rich.panel import Panel
from rich.text import Text
from textual import events
from textual.reactive import Reactive
from textual.widget import Widget

from .. import styles
from ..renderables.logrecord import LogRecord

PANEL_SIZE = 10


class Records(Widget):

    height: Reactive[int | None] = Reactive(None)

    def __init__(self, height: int | None = None):
        super().__init__()
        self.height = height
        self.records: List[LogRecord] = [
            # LogRecord(level=logging.INFO,
            #           msg="The log messages will be displayed here as a list or table.")
        ]
        self._selected_idx: Optional[int] = None

    async def on_mount(self) -> None:
        # self.layout_size = PANEL_SIZE
        # self.layout_fraction = 1
        ...

    async def on_click(self, event: events.Click) -> None:

        if self._selected_idx is not None:
            self.records[self._selected_idx].selected = False

        idx = event.y - 1  # Records is a Panel with the header as the first line

        try:
            record = self.records[idx]
            self.app.record_info.set(record)
            record.selected = True
            self._selected_idx = idx
            self.refresh(repaint=True)
        except IndexError:
            pass

    def render(self) -> Panel:
        return Panel(
            self._generate_renderable(),
            title="[bold]Records[/]",
            border_style=styles.BORDER_FOCUSED,
            box=styles.BOX,
            title_align="left",
            padding=0,
            height=self.height,
        )

    def replace(self, records: List[LogRecord]):
        self.records = records
        self._selected_idx = None

    def update(self, records: List[LogRecord]):
        self.records.extend(records)
        self._selected_idx = None

    def _generate_renderable(self) -> ConsoleRenderable:
        text = Text(no_wrap=True)
        [
            text.append(record.get_text()).append("\n")
            for record in self.records
        ]
        return text



from rich.table import Table
from textual.keys import Keys

from textualog.unicodes import DOWN
from textualog.unicodes import LEFT
from textualog.unicodes import RIGHT
from textualog.unicodes import UP


class Shortcuts:
    shortcuts = {
        "navigation": {
            "help": "?",
            "navigate": f"{LEFT} {RIGHT} {UP} {DOWN}",
            "close dialog": Keys.Escape,
            "quit": f"{Keys.ControlC} or q",
            "Show Namespaces": "n",
        },
        "Toggle Logging Levels": {
            "DEBUG mode": "d",
            "INFO mode": "i",
            "WARNING mode": "w",
            "ERROR mode": "e",
            "CRITICAL mode": "c",
        },
    }

    def __str__(self) -> str:
        return str(self.shortcuts)

    def __rich__(self) -> Table:
        table = Table(box=None, expand=False, show_footer=False, show_header=False)
        table.add_column(style="magenta bold")
        table.add_column(style="yellow bold")
        for category, shortcuts in self.shortcuts.items():
            table.add_row(f"[blue bold]{category}[/]")
            for action, shortcut in shortcuts.items():
                table.add_row(f"{action}:", f"{shortcut}")
            table.add_row()

        return table



import argparse
import sys
from pathlib import Path

from textual import events
from textual.app import App
from textual.keys import Keys
from textual.reactive import Reactive
from textual.widgets import Header

from . import __version__
from .loader import KeyValueLoader
from .renderables.namespace_tree import EntryClick
from .widgets.footer import Footer
from .widgets.help import Help
from .widgets.levels import Levels
from .widgets.namespaces import Namespaces
from .widgets.recordinfo import RecordInfo
from .widgets.records import Records


class TextualLog(App):

    show_help = Reactive(False)
    show_namespaces = Reactive(False)

    # The namespace_tree is just for demonstration purposes. The namespace should be a
    # tree like structure with proper navigation and the possibility to add and remove nodes.

    namespace_tree = {
        "egse": {
            "system": "system",
            "decorators": {"x": 1, "y": 2},
        }
    }

    def __init__(self, filename: str = None, **kwargs):
        super().__init__(**kwargs)
        self.filename = filename
        self.cursor = 0
        self.loader = None

    async def on_mount(self, event: events.Mount) -> None:
        """
        Call after terminal goes in to application mode.
        """

        self.namespaces = Namespaces("Name space", self.namespace_tree)
        self.namespaces.layout_offset_x = -40

        self.help_widget = Help()
        self.help_widget.visible = False

        await self.view.dock(Header(), edge="top")
        await self.view.dock(Footer(), edge="bottom")
        await self.view.dock(self.namespaces, edge="left", size=40, z=1)
        await self.view.dock(self.help_widget, edge="right", size=40, z=1)
        grid = await self.view.dock_grid(edge="left", name="left")

        grid.add_column(size=30, name="left")
        grid.add_column(fraction=1, name="right", min_size=100)

        grid.add_row(fraction=1, name="top")
        grid.add_row(fraction=1, name="middle")
        grid.add_row(size=7, name="bottom")

        grid.add_areas(
            area1="left-start|right-end,top-start|middle-end",
            area2="left,bottom",
            area3="right,bottom",
        )

        self.levels = Levels()
        self.records = Records()
        self.record_info = RecordInfo()

        grid.place(
            area1=self.records,
            area2=self.levels,
            area3=self.record_info,
        )

        if self.filename:
            self.loader = KeyValueLoader(self.filename)
            self.loader.load()
            self.loader.process(0, 500, None)

            # The height of the self.records view is not yet known, so we take a large enough number

            self.records.update(self.loader.get_records(0, 500, None))

    async def on_load(self) -> None:
        """
        Sent before going in to application mode. This method is called before on_mount().
        Bind keys here.
        """
        await self.bind("q", "quit", "Quit")
        await self.bind("?", "toggle_help", "Help")

    async def on_key(self, event) -> None:

        if self.loader is None:
            return

        # The height of the text area of the Records panel

        height = self.records.size.height - 2
        size = self.loader.size()

        self.app.sub_title = f"Key pressed: {event.key}"

        if event.key == "d":
            self.levels.debug_level = not self.levels.debug_level
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))
        elif event.key == "i":
            self.levels.info_level = not self.levels.info_level
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))
        elif event.key == "w":
            self.levels.warning_level = not self.levels.warning_level
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))
        elif event.key == "e":
            self.levels.error_level = not self.levels.error_level
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))
        elif event.key == "c":
            self.levels.critical_level = not self.levels.critical_level
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))
        elif event.key in "nN":
            self.show_namespaces = not self.show_namespaces
        elif event.key == "r":
            self.loader.load()
        elif event.key == Keys.Escape:
            self.show_help = False
            self.show_namespaces = False
        elif event.key == Keys.Down:
            self.cursor = min(size, self.cursor + 1)
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))
        elif event.key == Keys.Up:
            self.cursor = max(0, self.cursor - 1)
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))
        elif event.key == Keys.PageDown:
            self.cursor = min(size, self.cursor+(height-1))
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))
        elif event.key == Keys.PageUp:
            self.cursor = max(0, self.cursor-(height-1))
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))
        elif event.key == Keys.End:
            self.cursor = size - height
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))
        elif event.key == Keys.Home:
            self.cursor = 0
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))

        self.records.refresh(layout=True, repaint=True)

    async def watch_show_namespaces(self, show_namespaces: bool) -> None:
        """Called when show_namespaces changes."""
        self.namespaces.animate("layout_offset_x", 0 if show_namespaces else -40)

    async def watch_show_help(self, show_help: bool) -> None:
        """Called when show_help changes."""
        self.help_widget.visible = show_help

    async def action_toggle_help(self) -> None:
        """Called when bound help key is pressed."""
        self.show_help = not self.show_help

    async def handle_entry_click(self, message: EntryClick) -> None:
        """A message sent by the namespace tree when an entry is clicked."""

        self.app.sub_title = f"{message.key}"
        self.records.refresh(layout=True)


def main():
    from rich.traceback import install
    install(show_locals=False)

    parser = argparse.ArgumentParser(
        description="Textual Log Viewer, display, filter and search log files",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "--version",
        "-v",
        action="version",
        version=_get_version_text(),
        help="display version information",
    )

    parser.add_argument(
        "--log",
        "-l",
        type=str,
        default=None,
        help="debug log file",
    )

    args = parser.parse_args()

    if args.log and not Path(args.log).exists():
        raise FileNotFoundError(f"No such file {args.log}")

    TextualLog.run(title="Textual Log Viewer", log=None, filename=args.log)


def _get_version_text():
    python_version = f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"

    return "\n".join(
        [
            f"textualog {__version__.__version__} [Python {python_version}]",
            "Copyright © 2022 Rik Huygen",
        ]
    )


if __name__ == "__main__":
    main()



from rich.console import RenderableType
from rich.panel import Panel

from textualog import styles
from textualog.renderables.namespace_tree import NamespaceTree


class Namespaces(NamespaceTree):
    def __init__(self, name=None, data: dict = None):
        self.tree: dict = data
        super().__init__(name=name, data=data)

    def render(self) -> RenderableType:
        panel = Panel(
            super().render(),
            title="[bold]Namespaces[/]",
            border_style=styles.BORDER,
            box=styles.BOX,
            title_align="left",
            padding=0,
        )
        return panel



from __future__ import annotations

from dataclasses import dataclass
from functools import lru_cache
from typing import Union

import rich
from rich.console import RenderableType
from rich.text import Text
from textual import events
from textual._types import MessageTarget
from textual.message import Message
from textual.reactive import Reactive
from textual.widgets import NodeID
from textual.widgets import TreeClick
from textual.widgets import TreeControl
from textual.widgets import TreeNode


@dataclass
class NamespaceEntry:
    name: dict
    is_parent: bool

    def __str__(self):
        return f"{type(self.name)=}, {self.name=}, {self.is_parent=}"


@rich.repr.auto
class EntryClick(Message, bubble=True):
    def __init__(self, sender: MessageTarget, key: Union[dict, str]) -> None:
        self.key = key
        super().__init__(sender)


class NamespaceTree(TreeControl[NamespaceEntry]):
    def __init__(self, data: dict = None, name: str = None):
        label = "log"
        data = NamespaceEntry(data, True)
        super().__init__(label, name=name, data=data)
        self.root.tree.guide_style = "black"

    has_focus: Reactive[bool] = Reactive(False)
    mouse_over: Reactive[bool] = Reactive(False)
    style: Reactive[str] = Reactive("")
    height: Reactive[int | None] = Reactive(None)

    def on_focus(self) -> None:
        self.has_focus = True

    def on_blur(self) -> None:
        self.has_focus = False

    async def watch_hover_node(self, hover_node: NodeID) -> None:
        for node in self.nodes.values():
            node.tree.guide_style = (
                "bold not dim red" if node.id == hover_node else "black"
            )
        self.refresh(layout=True)

    def render_node(self, node: TreeNode[NamespaceEntry]) -> RenderableType:
        return self.render_tree_label(
            node,
            node.data.is_parent,
            node.expanded,
            node.is_cursor,
            node.id == self.hover_node,
            self.has_focus,
        )

    @lru_cache(maxsize=1024 * 32)
    def render_tree_label(
        self,
        node: TreeNode[NamespaceEntry],
        is_parent: bool,
        expanded: bool,
        is_cursor: bool,
        is_hover: bool,
        has_focus: bool,
    ) -> RenderableType:
        meta = {
            "@click": f"click_label({node.id})",
            "tree_node": node.id,
            "cursor": node.is_cursor,
        }
        label = Text(node.label) if isinstance(node.label, str) else node.label
        if is_hover:
            label.stylize("underline")
        if is_parent:
            label.stylize("bold magenta")
            icon = "▼︎" if expanded else "▶︎"
        else:
            label.stylize("bright_green")
            icon = ""
            label.highlight_regex(r"\..*$", "green")

        if label.plain.startswith("."):
            label.stylize("dim")

        if is_cursor and has_focus:
            label.stylize("reverse")

        icon_label = Text(f"{icon} ", no_wrap=True, overflow="ellipsis") + label
        icon_label.apply_meta(meta)
        return icon_label

    async def on_mount(self, event: events.Mount) -> None:
        await self.load_tree(self.root)

    async def load_tree(self, node: TreeNode[NamespaceEntry]):
        try:
            keys = sorted(node.data.name.keys())
        except AttributeError:
            keys = []
        for key in keys:
            is_dir = isinstance(node.data.name[key], dict)
            await node.add(key, NamespaceEntry(node.data.name[key], is_dir))
        node.loaded = True
        await node.expand()
        self.refresh(layout=True)

    async def handle_tree_click(self, message: TreeClick[NamespaceEntry]) -> None:
        namespace_entry = message.node.data
        if not namespace_entry.is_parent:
            await self.emit(EntryClick(self, namespace_entry.name))
        else:
            await self.emit(EntryClick(self, namespace_entry.name))
            if not message.node.loaded:
                await self.load_tree(message.node)
                await message.node.expand()
            else:
                await message.node.toggle()



from __future__ import annotations

from typing import Optional

from rich.panel import Panel
from rich.text import Text
from textual.reactive import Reactive
from textual.widget import Widget

from .. import styles
from ..renderables.logrecord import LogRecord

PANEL_SIZE = 5


class RecordInfo(Widget):

    record: Reactive[LogRecord] = Reactive(None)

    def __init__(self, height: int | None = None):
        super().__init__()
        self.record: Optional[LogRecord] = None

    async def on_mount(self) -> None:
        # self.layout_size = PANEL_SIZE
        # self.layout_fraction = 1
        ...

    def set(self, record: LogRecord):
        self.record = record

    def render(self) -> Panel:

        return Panel(
            self._generate_renderable(),
            title="[bold]Record Info[/]",
            border_style=styles.BORDER_FOCUSED,
            box=styles.BOX,
            title_align="left",
            padding=0,
            # height=PANEL_SIZE,
        )

    def _generate_renderable(self) -> Text:

        if self.record is None:
            return Text()

        record = Text(no_wrap=True)
        record.append(f"level      = {self.record.level}\n")
        record.append(f"process    = {self.record.process}\n")
        record.append(f"process ID = {self.record.process_id}\n")
        record.append(f"caller     = {self.record.caller}\n")
        record.append(f"msg        = {self.record.msg}\n")

        return record
